# -*- coding: utf-8 -*-

# Copyright (C) 2014 - 2020 Honda Research Institute Japan Co., Ltd.
# All rights reserved.
# HARK was developed by researchers in Okuno Laboratory
# at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
# 
# Licensed under the HARK license, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# https://www.hark.jp/notice/HARK_License_Agreement.pdf

import sys
import xml.etree.ElementTree as ET
from xml.dom import minidom

def prettify(elem):
    #Return a pretty-printed XML string for the Element.
    rough_string = ET.tostring(elem, 'utf-8')
    reparsed = minidom.parseString(rough_string)
    return reparsed.toprettyxml(indent="    ")

argv = sys.argv
if len(argv) < 2:
    print("Usage: " + argv[0] + " input.xml [output.xml]")
    sys.exit()
inputxml = argv[1]
outputxml = argv[2] if len(argv) >= 3 else "output.xml"

# Load xml and validation
orgTree = ET.parse(inputxml)
orgRoot = orgTree.getroot()
if orgRoot.tag != "hark_xml":
    print("Input file is not HARK xml!")
    sys.exit()
orgConfig = orgRoot.find("./hark_config")
if orgConfig is None or orgConfig.get("major") != "2" or orgConfig.get("minor") != "0":
    print("Unspported HARK xml format!")
    sys.exit()
srcType = orgConfig.get("id")
if srcType is None or ( srcType != "TSPList" and srcType != "ImpulseResponseList" ):
    print("Unspported source id!")
    sys.exit()

""" Process conversion
"""
newRoot = ET.Element("hark_xml", {'version': '1.3'})
newConfig = ET.SubElement(newRoot, 'config')

# config
ET.SubElement(newConfig, "comment").text = "conversion HARK xml from libharkio2 to libharkio3"
if srcType == "TSPList":
    ET.SubElement(newConfig, "SynchronousAverage").text = orgConfig.findtext("./SynchronousAverage", "")
    ET.SubElement(newConfig, "TSPpath").text = "/usr/share/hark/16384.little_endian.wav"
    ET.SubElement(newConfig, "TSPOffset").text = orgConfig.findtext("./TSPOffset", "")
    ET.SubElement(newConfig, "samplingRate").text = "16000"
    ET.SubElement(newConfig, "signalMax").text = orgConfig.findtext("./SignalMax", "")
    ET.SubElement(newConfig, "TSPLength").text = orgConfig.findtext("./TSPLength", "")
ET.SubElement(newConfig, "PeakSearch", {'from': '0', 'to': '0'})
ET.SubElement(newConfig, "nfft").text = "512"

# position
if srcType == "ImpulseResponseList":
    ft = open("flt2multiwav.txt", "wb")

positions = ET.SubElement(newRoot, "positions", {'type': 'tsp' if srcType == "TSPList" else 'impulse', 'coordinate': 'cartesian'})
id = 0
channels = ""
for e in orgConfig.findall("./Sources/Source"):
    pos = e.find("Position")
    if pos is not None:
        if pos.get("type") != "cartesian":
            print("Unsupported position coordinate!")
            sys.exit()
        pos_x = (pos.findtext("X", ""))
        pos_y = (pos.findtext("Y", ""))
        pos_z = (pos.findtext("Z", ""))

        if channels == "":
            mi = e.findall("./Files/File/MicIndices/MicIndex")
            if mi is None:
                channels = "all"
            else:
                for m in mi:
                    channels += (m.text + ';')
                if channels == "0;1;2;3;4;5;6;7;":
                    channels = "all"

        if srcType == "TSPList":
            path = e.findtext("./Files/File/Name", "")
        elif srcType == "ImpulseResponseList":
            path = "impulse_response_" + '%04d' % id + ".wav"
            fltfiles = e.findall("./Files/File/Name")
            flt2multiwavComm = "harktoolcli-flt2multiwav " + path
            for flt in fltfiles:
                flt2multiwavComm += (" " + flt.text)
            ft.write(flt2multiwavComm + '\n')
        else:
            path = ""

        ET.SubElement(positions, "position", {'x': pos_x, 'y': pos_y, 'z': pos_z, 'id': str(id), 'path': path})
    id+=1

if srcType == "ImpulseResponseList":
    ft.close()

# channel
if channels != "" and channels != "all":
    ET.SubElement(newRoot, "channels", {'use': channels})

# Write
with open(outputxml, "wb") as f:
    f.write(prettify(newRoot))
