/*
 * Copyright (c) 2014 - 2020 Honda Research Institute Japan Co., Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#include "harktool.h"
/*-----------------------------------------------------------------------------
*  FFT 計算       (float complex --> float complex)
*
*  COMPLEX_FLOAT*  a  // 入力列
*  COMPLEX_FLOAT*  A  // 出力・計算用配列（in-place計算）
*  int fftsize        // FFT長
-----------------------------------------------------------------------------*/
void harktool_FFTImpl_fFFT(const COMPLEX_FLOAT *a, COMPLEX_FLOAT *A, int fftsize)
{
    // 初期化
    if (a != A) {
        memcpy((void*) A, (void*) a, fftsize * sizeof(COMPLEX_FLOAT));
    }
    COMPLEX_FLOAT w[fftsize];
    for (int i = 0; i < fftsize; i++) {
        w[i] = cos(two_pi * i / fftsize) - sin(two_pi * i / fftsize) * _Complex_I;
    }

    // ビット反転並べ替え
    int i = 0;
    for (int j = 1; j < fftsize - 1; j++) {
        for (int k = fftsize >> 1; k > (i ^= k); k >>= 1)
            ;
        if (j < i) {
            harktool_swapFloatComplex(&A[i], &A[j]);
        }
    }

    // 主要計算部
    int m;
    for (int mh = 1; (m = mh << 1) <= fftsize; mh = m) {
        int irev = 0;
        for (int i = 0; i < fftsize; i += m) {
            int prev_irev = irev;
            // w[prev_irev] = cos(two_pi * irev / fftsize) - sin(two_pi * irev / fftsize) * _Complex_I;
            int k;
            for (k = fftsize >> 2; k > (irev ^= k); k >>= 1)
                ;
            for (int j = i; j < mh + i; j++) {
                k = j + mh;
                COMPLEX_FLOAT y = A[j] - A[k];
                A[j] = A[j] + A[k];
                A[k] = w[prev_irev] * y;
            }
        }
    }
}

/*-----------------------------------------------------------------------------
*  FFT 逆計算       (float complex <-- float complex)
*
*  COMPLEX_FLOAT*  A  // 入力列
*  COMPLEX_FLOAT*  a  // 出力・計算用配列（in-place計算）
*  int fftsize        // FFT長
-----------------------------------------------------------------------------*/
void harktool_FFTImpl_fFFTInv(const COMPLEX_FLOAT *A, COMPLEX_FLOAT *a, int fftsize)
{
    // 初期化
    if (A != a) {
        memcpy((void*) a, (void*) A, fftsize * sizeof(COMPLEX_FLOAT));
    }
    COMPLEX_FLOAT w[fftsize];
    for (int i = 0; i < fftsize; i++) {
        w[i] = cos(two_pi * i / fftsize) - sin(two_pi * i / fftsize) * _Complex_I;
    }

    // ビット反転並べ替え
    int i = 0;
    for (int j = 1; j < fftsize - 1; j++) {
        for (int k = fftsize >> 1; k > (i ^= k); k >>= 1)
            ;
        if (j < i) {
            harktool_swapFloatComplex(&a[i], &a[j]); // std::swap(a[i], a[j]);
        }
    }

    // 主要計算部
    int m;
    for (int mh = 1; (m = mh << 1) <= fftsize; mh = m) {
        int irev = 0;
        for (int i = 0; i < fftsize; i += m) {
            int prev_irev = irev;
            // w[prev_irev] = cos(two_pi * irev / fftsize) - sin(two_pi * irev / fftsize) * _Complex_I;
            int k;
            for (k = fftsize >> 2; k > (irev ^= k); k >>= 1)
                ;
            for (int j = i; j < mh + i; j++) {
                k = j + mh;
                COMPLEX_FLOAT y = a[j] - a[k];
                a[j] = a[j] + a[k];
                a[k] = conjf(w[prev_irev]) * y;
            }
        }
    }
}

/*-----------------------------------------------------------------------------
*  実FFT 順計算       (float 実数 --> float 複素共役対称列)
*
*  float*  float_a    // 入力実数配列（time, 長さ n）
*  COMPLEX_FLOAT*  A  // 出力複素配列 (freq, 長さ n/2)
*  int fftsize        // FFT長 n
-----------------------------------------------------------------------------*/
void harktool_FFTImpl_fRealFFT(const float *float_a, COMPLEX_FLOAT *A, int fftsize) {
    COMPLEX_FLOAT complex_a[fftsize];
    for (int i = 0; i < fftsize; i++) {
        complex_a[i] = float_a[i] + 0.0 * _Complex_I;
    }

    harktool_FFTImpl_fFFT(complex_a, A, fftsize);
}

/*-----------------------------------------------------------------------------
*  実FFT 逆計算       (float 実数 <-- float 複素共役対称列)
*
*  COMPLEX_FLOAT*  A  // 入力複素配列 (freq, 長さ n/2+1)
*  float*  float_a    // 出力実数配列（time, 長さ n）
*  int fftsize        // FFT長 n  ( n > 4 )
-----------------------------------------------------------------------------*/
void harktool_FFTImpl_fRealFFTInv(const COMPLEX_FLOAT *A, float *float_a, int fftsize)
{
    const int n = fftsize;
    const int n_2 = n / 2;
    const int n_4 = n / 4;

    COMPLEX_FLOAT *tmp_A; // 半サイズのFFT作業用

    tmp_A = (COMPLEX_FLOAT *)malloc((n_2) * sizeof(COMPLEX_FLOAT));
    if (tmp_A == NULL) {
        harkio_Log_printf(harkio_Log_Error, "malloc error");
        exit(EXIT_FAILURE);
    }

    // 変数変換
    tmp_A[0] = (crealf(A[0]) + crealf(A[n_2])) + (crealf(A[0]) - crealf(A[n_2])) * _Complex_I;
    tmp_A[0] = 0.5f * tmp_A[0]; // A[n_2]はここでのみ使う
    tmp_A[n_4] = conjf(A[n_4]);

    for (int k = 1; k < n_4; k++) {
        float w_k_real = cosf(two_pi * k / n);
        float w_k_imag = -sinf(two_pi * k / n);

        COMPLEX_FLOAT z = (1.0 + w_k_imag) + w_k_real * _Complex_I; // 1 + i*conj(Wk)
        COMPLEX_FLOAT y = 0.5 * z * (A[k] - conjf(A[n_2 - k]));

        tmp_A[k] = conjf(A[n_2 - k]) + y;
        tmp_A[n_2 - k] = conjf(A[k] - y);
    }

    // 長さ半分の複素IFFTを実行
    harktool_FFTImpl_fFFTInv(tmp_A, tmp_A, n_2);

    // 変数変換
    for (int i = 0; i < n_2; i++) {
        float_a[i * 2] = crealf(tmp_A[i]) * 2;
        float_a[i * 2 + 1] = cimagf(tmp_A[i]) * 2;
    }
    free(tmp_A);
}

/*-----------------------------------------------------------------------------
*  FFT 計算       (double complex --> double complex)
*
*  COMPLEX_DOUBLE*  a  // 入力列
*  COMPLEX_DOUBLE*  A  // 出力・計算用配列（in-place計算）
*  int fftsize         // FFT長
-----------------------------------------------------------------------------*/
void harktool_FFTImpl_dFFT(const COMPLEX_DOUBLE *a, COMPLEX_DOUBLE *A, int fftsize)
{
    // 初期化
    if (a != A) {
        memcpy((void*) A, (void*) a, fftsize * sizeof(COMPLEX_DOUBLE));
    }
    COMPLEX_DOUBLE w[fftsize];
    for (int i = 0; i < fftsize; i++) {
        w[i] = cos(two_pi * i / fftsize) - sin(two_pi * i / fftsize) * _Complex_I;
    }

    // ビット反転並べ替え
    int i = 0;
    for (int j = 1; j < fftsize - 1; j++) {
        for (int k = fftsize >> 1; k > (i ^= k); k >>= 1)
            ;
        if (j < i) {
            harktool_swapDoubleComplex(&A[i], &A[j]);
        }
    }

    // 主要計算部
    int m;
    for (int mh = 1; (m = mh << 1) <= fftsize; mh = m) {
        int irev = 0;
        for (int i = 0; i < fftsize; i += m) {
            int prev_irev = irev;
            // w[prev_irev] = cos(two_pi * irev / fftsize) - sin(two_pi * irev / fftsize) * _Complex_I;
            int k;
            for (k = fftsize >> 2; k > (irev ^= k); k >>= 1)
                ;
            for (int j = i; j < mh + i; j++) {
                k = j + mh;
                COMPLEX_DOUBLE y = A[j] - A[k];
                A[j] = A[j] + A[k];
                A[k] = w[prev_irev] * y;
            }
        }
    }
}

/*-----------------------------------------------------------------------------
*  FFT 逆計算       (double complex <-- double complex)
*
*  COMPLEX_DOUBLE*  A  // 入力列
*  COMPLEX_DOUBLE*  a  // 出力・計算用配列（in-place計算）
*  int fftsize         // FFT長
-----------------------------------------------------------------------------*/
void harktool_FFTImpl_dFFTInv(const COMPLEX_DOUBLE *A, COMPLEX_DOUBLE *a, int fftsize)
{
    // 初期化
    if (A != a) {
        memcpy((void*) a, (void*) A, fftsize * sizeof(COMPLEX_DOUBLE));
    }
    COMPLEX_DOUBLE w[fftsize];
    for (int i = 0; i < fftsize; i++) {
        w[i] = cos(two_pi * i / fftsize) - sin(two_pi * i / fftsize) * _Complex_I;
    }

    // ビット反転並べ替え
    int i = 0;
    for (int j = 1; j < fftsize - 1; j++) {
        for (int k = fftsize >> 1; k > (i ^= k); k >>= 1)
            ;
        if (j < i) {
            harktool_swapDoubleComplex(&a[i], &a[j]); // std::swap(a[i], a[j]);
        }
    }

    // 主要計算部
    int m;
    for (int mh = 1; (m = mh << 1) <= fftsize; mh = m) {
        int irev = 0;
        for (i = 0; i < fftsize; i += m) {
            int prev_irev = irev;
            // w[prev_irev] = cos(two_pi * irev / fftsize) - sin(two_pi * irev / fftsize) * _Complex_I;
            int k;
            for (k = fftsize >> 2; k > (irev ^= k); k >>= 1)
                ;
            for (int j = i; j < mh + i; j++) {
                k = j + mh;
                COMPLEX_DOUBLE y = a[j] - a[k];
                a[j] = a[j] + a[k];
                a[k] = conj(w[prev_irev]) * y;
            }
        }
    }
}

/*-----------------------------------------------------------------------------
*  実FFT 順計算       (double 実数 --> double 複素共役対称列)
*
*  double*  double_a    // 入力実数配列（time, 長さ n）
*  COMPLEX_DOUBLE*  A  // 出力複素配列 (freq, 長さ n/2)
*  int fftsize         // FFT長 n
-----------------------------------------------------------------------------*/
void harktool_FFTImpl_dRealFFT(const double *double_a, COMPLEX_DOUBLE  *A, int fftsize) {
    COMPLEX_DOUBLE complex_a[fftsize];
    for (int i = 0; i < fftsize; i++) {
        complex_a[i] = double_a[i] + 0.0 * _Complex_I;
    }

    harktool_FFTImpl_dFFT(complex_a, A, fftsize);
}

/*-----------------------------------------------------------------------------
*  実FFT 逆計算       (double 実数 <-- double 複素共役対称列)
*
*  COMPLEX_DOUBLE*  A  // 入力複素配列 (freq, 長さ n/2+1)
*  double*  double_a   // 出力実数配列（time, 長さ n）
*  int fftsize         // FFT長 n  ( n > 4 )
*  COMPLEX_DOUBLE*  w  // 作業用配列（NULLの場合は関数内で確保する）
-----------------------------------------------------------------------------*/
void harktool_FFTImpl_dRealFFTInv(const COMPLEX_DOUBLE *A, double *double_a, int fftsize)
{
    const int n = fftsize;
    const int n_2 = n / 2;
    const int n_4 = n / 4;

    COMPLEX_DOUBLE *tmp_A; // 半サイズのFFT作業用

    tmp_A = (COMPLEX_DOUBLE *)malloc((n_2) * sizeof(COMPLEX_DOUBLE));
    if (tmp_A == NULL) {
        harkio_Log_printf(harkio_Log_Error, "malloc error");
        exit(EXIT_FAILURE);
    }

    // 変数変換
    tmp_A[0] = (creal(A[0]) + creal(A[n_2])) + (creal(A[0]) - creal(A[n_2])) * _Complex_I;
    tmp_A[0] = 0.5f * tmp_A[0]; // A[n_2]はここでのみ使う
    tmp_A[n_4] = conj(A[n_4]);

    for (int k = 1; k < n_4; k++) {
        double w_k_real = cos(two_pi * k / n);
        double w_k_imag = -sin(two_pi * k / n);

        COMPLEX_DOUBLE z = (1.0 + w_k_imag) + w_k_real * _Complex_I; // 1 + i*conj(Wk)
        COMPLEX_DOUBLE y = 0.5 * z * (A[k] - conj(A[n_2 - k]));

        tmp_A[k] = conj(A[n_2 - k]) + y;
        tmp_A[n_2 - k] = conj(A[k] - y);
    }

    // 長さ半分の複素IFFTを実行
    harktool_FFTImpl_dFFTInv(tmp_A, tmp_A, n_2);

    // 変数変換
    for (int i = 0; i < n_2; i++) {
        double_a[i * 2] = creal(tmp_A[i]) * 2;
        double_a[i * 2 + 1] = cimag(tmp_A[i]) * 2;
    }
}

/*----------------------------------------------------------------*
* float complex型　swap
*
-----------------------------------------------------------------*/
void harktool_swapFloatComplex(float _Complex  *a , float _Complex  *b)
{
    float _Complex tmp = *a;
    *a = *b;
    *b = tmp;
}

/*----------------------------------------------------------------*
* double complex型　swap
*
-----------------------------------------------------------------*/
void harktool_swapDoubleComplex(double _Complex  *a , double _Complex  *b)
{
    double _Complex tmp = *a;
    *a = *b;
    *b = tmp;
}
