/*
 * Copyright (c) 2014 - 2020 Honda Research Institute Japan Co., Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

/*-----------------------------------------
 %
 % FD
 %
 % Algorithm ---------------------------
 % X_1 = A * S_1
 % X_2 = A * S_2
 % X_3 = A * S_3
 %...
 % X_N = A * S_N
 % Sit = [S_1, S_2, S_3, ..., S_N]'
 % Sot = [X_1, X_2, X_3, ..., X_N]'
 % hatA = inv(Sit' * Sit) * Sit' * Sot
 %



 #  Arguments ---------------------------
 #  ARG01(src)        : src number [Integer]
 #  ARG02(inFile)     : Input wav file name [String]
 #  ARG03(outFile)    : Output flt filename prefix. The name will be [outDir outFile '_ch*.flt']  [String]
 #  ARG04(s)          : start point for Transfer Function(TF) calculation in sample [Integer]
 #  ARG05(e)          : end point for TF calculation in sample [Integer]
 #  ARG06(nch)        :  Channel Total number  [Integer]
 #  ARG07(fltlen)     : Number of samples output as TF [Integer]
 #  ARG08(nfft)       : Number of FFT points [Integer]
 #  ARG09(hop)        : Shift length for TF calculation [Integer]
 #  ARG10(headmargin) : Head margin of impulse responces in samples [Integer(Default:0)]
 #  ARG11(eneble_wn)  : Enables adding white noise to escape unstable zero signal [Bool]
 #  ARG12(wn_level)   : Level of white noise. Set enable_wn=true if you want to use this. [Double]
 #  ARG13(enable_normalize) : Enable normalizing output TF
 ------------------------------------------------------------------------------------*/

#include "harktool.h"

//プロトタイプ宣言
void CalcHammingWindow(double *w, int length);

int harktool_mouthTspFD(const char *inFile, // Input wav file name [String]
        const char *outFile, // Output flt filename prefix. The name will be [outDir outFile '_ch*.flt']  [String]
        int s, // start point for Transfer Function(TF) calculation in sample [Integer]
        int e, // end point for TF calculation in sample [Integer]
        int mch, // Channel index of the voice input. [Integer]
        const bool* outChannels, // Select channels for output.(MAX_CH)
        int fltlen, // Number of samples output as TF [Integer]
        int nfft, // Number of FFT points [Integer]
        int hop, // Shift length for TF calculation [Integer]
        int headmargin, // Head margin of impulse responces in samples [Integer(Default:0)]
        bool enable_wn, // Enables adding white noise to escape unstable zero signal [bool]
        double wn_level, // Level of white noise. Set enable_wn=true if you want to use this. [Double]
        int cfgSamplingRate    // Sampling rate of configuration
        ) {
    // Print mouthTspFD
    char* msg = NULL;
    cat_asprintf(&msg, "\nharktool_mouthTspFD   inFile         [%s]\n", inFile);
    cat_asprintf(&msg, "                    , outFile        [%s]\n", outFile);
    cat_asprintf(&msg, "                    , s              [%d]\n", s);
    cat_asprintf(&msg, "                    , e              [%d]\n", e);
    cat_asprintf(&msg, "                    , mch            [%d]\n", mch);
    cat_asprintf(&msg, "                    , outputChannels [");
    bool printedCh = false;
    for (int i = 0; i < MAX_CH; i++) {
        if (outChannels[i]) {
            if (printedCh)
                cat_asprintf(&msg, ",");
            printedCh = true;
            cat_asprintf(&msg, "%d", i);
        }
    }
    cat_asprintf(&msg, "]\n");
    cat_asprintf(&msg, "                    , fltlen         [%d]\n", fltlen);
    cat_asprintf(&msg, "                    , nfft           [%d]\n", nfft);
    cat_asprintf(&msg, "                    , hop            [%d]\n", hop);
    cat_asprintf(&msg, "                    , headmargin     [%d]\n", headmargin);
    cat_asprintf(&msg, "                    , enable_wn      [%s]\n", enable_wn);
    cat_asprintf(&msg, "                    , wn_level       [%lf]\n", wn_level);
    cat_asprintf(&msg, "                    , samplingrate   [%d]\n", cfgSamplingRate);
    harkio_Log_printf(harkio_Log_Info, "%s", msg);
    free(msg);

    // Load wave file
    double **tsp_wave;
    int datasize;
    int channelnum;
    int samplingrate;
    bool isParamOK = true;
    if (readwavfile(inFile, &tsp_wave, &datasize, &channelnum, &samplingrate) == -1) {
        harkio_Log_printf(harkio_Log_Error, "wavefile(%s) is not found.", inFile);
        isParamOK = false;
    }
    // parameter check from wav file
    if (datasize < e) {
        harkio_Log_printf(harkio_Log_Error, "Wav(%s) frames(%d) is less than end of sample (%d)", datasize, inFile, e);
        isParamOK = false;
    }
    if (mch >= channelnum) {
        harkio_Log_printf(harkio_Log_Error, "mouth TSP data not found: wav channel [%d] in file [%s]", mch, inFile);
        isParamOK = false;
    }
    if (cfgSamplingRate != samplingrate) {
        harkio_Log_printf(harkio_Log_Error, "Config SamplingRate (%d) differ from Wav(%s) SamplingRate (%d)",
                cfgSamplingRate, inFile, samplingrate);
        isParamOK = false;
    }
    if (!isParamOK) {
        // free wave buffer
        for (int i = 0; i < channelnum; i++) {
            free(tsp_wave[i]);
        }
        free(tsp_wave);
        return EXIT_FAILURE;
    }

    /*--------------------------------------------------*/
    // Input
    /*--------------------------------------------------*/
    //	mch の s - e の抜き出し
    int input_org_size = e - s;
    double* input_org = &tsp_wave[mch][s];

    double window[nfft];
    CalcHammingWindow(window, nfft);

    // fft
    int numShift = (input_org_size - nfft) / hop + 1;
    COMPLEX_DOUBLE** Si = harktool_newComplexDoubleTwoDim(numShift, nfft / 2 + 1);
    for (int j = 0; j < numShift; j++) {
        COMPLEX_DOUBLE spec[nfft];
        for (int k = 0; k < nfft; k++) {
            spec[k] = window[k] * input_org[j * hop + k];
        }

        harktool_FFTImpl_dFFT(spec, spec, nfft);

        for (int n = 0; n < nfft / 2 + 1; n++) {
            Si[j][n] = spec[n];
        }
    }

    /*--------------------------------------------------*/
    // Output
    /*--------------------------------------------------*/
    //int output_org_size = e - s;
    int nch = 0;
    int outChIds[MAX_CH];
    // calc nch and outChIds[]
    for (int i = 0; i < MAX_CH; i++) {
        if (outChannels[i]) {
            // check channel
            if (i >= channelnum) {
                harkio_Log_printf(harkio_Log_Error, "mouth TSP data not found: wav channel [%d] in file [%s]", i,
                        inFile);
                // free Si
                harktool_freeComplexDoubleTwoDim(Si);
                // free wave buffer
                for (int i = 0; i < channelnum; i++) {
                    free(tsp_wave[i]);
                }
                free(tsp_wave);
                return EXIT_FAILURE;
            }
            outChIds[nch] = i;
            nch++;
        }
    }

    for (int i = 0; i < nch; i++) {
        // (s-headmargin) - (e-headmargin) の抜き出し
        double* output_org = &tsp_wave[outChIds[i]][s - headmargin];

        COMPLEX_DOUBLE** So = harktool_newComplexDoubleTwoDim(numShift, nfft / 2 + 1);
        for (int j = 0; j < numShift; j++) {
            COMPLEX_DOUBLE spec[nfft];
            for (int k = 0; k < nfft; k++) {
                spec[k] = window[k] * output_org[j * hop + k];
            }

            harktool_FFTImpl_dFFT(spec, spec, nfft);

            //for (n = 0; n < nfft; n++) {
            //    complexWave[n].im = complexWave[n].im;// * (-1);
            //}

            for (int n = 0; n < nfft / 2 + 1; n++) {
                So[j][n] = spec[n];
            }
        }

        COMPLEX_DOUBLE Ro[nfft];
        int j;
        for (j = 0; j < nfft / 2 + 1; j++) {
            COMPLEX_DOUBLE Sit[numShift][1];
            COMPLEX_DOUBLE Sit_[1][numShift];
            COMPLEX_DOUBLE Sot[numShift][1];
            for (int n = 0; n < numShift; n++) {
                Sit[n][0] = Si[n][j];
                Sit_[0][n] = conj(Sit[n][0]);
                Sot[n][0] = So[n][j];
            }

            // inv(Sit' * Sit)
            double invSit_Sit = 0.0f;
            for (int n = 0; n < numShift; n++) {
                invSit_Sit += creal(Sit_[0][n] * Sit[n][0]);
            }
            invSit_Sit = 1 / invSit_Sit;

            // Sit' * Sot
            COMPLEX_DOUBLE Sit_Sot = 0.0;
            for (int n = 0; n < numShift; n++) {
                Sit_Sot += Sit_[0][n] * Sot[n][0];
            }
            Ro[j] = invSit_Sit * Sit_Sot;
        }
        for (; j < nfft; j++) {
            Ro[j] = 0.0;
        }

        harktool_FFTImpl_dFFTInv(Ro, Ro, nfft);

        float YY[nfft];
        for (int j = 0; j < nfft; j++) {
            YY[j] = creal(Ro[j]) / nfft;
        }

        // save impulse response files
        char* file_w = NULL;
        asprintf(&file_w, "%s_ch%d.flt", outFile, outChIds[i]);
        FILE *fp_w;
        if ((fp_w = fopen(file_w, "wb")) == NULL) {
            harkio_Log_FileOpenFailed(file_w);
            // free wave buffer
            for (int i = 0; i < channelnum; i++) {
                free(tsp_wave[i]);
            }
            free(tsp_wave);
            if (file_w)
                free(file_w);
            return EXIT_FAILURE;
        }

        for (int j = 0; j < fltlen; j++) { //fltlen=4096
            fwrite(&YY[j], sizeof(float), 1, fp_w);
        }
        if (file_w)
            free(file_w);
        fclose(fp_w);
        // free So
        harktool_freeComplexDoubleTwoDim(So);
    }

    // free Si
    harktool_freeComplexDoubleTwoDim(Si);
    // free wave buffer
    for (int i = 0; i < channelnum; i++) {
        free(tsp_wave[i]);
    }
    free(tsp_wave);

    return EXIT_SUCCESS;
}

void CalcHammingWindow(double *w, int length) {
    int n;
    double fft_window_gain = 1.0;

    double c = 2.0 * M_PI / ((double) length - 1);

    for (n = 0; n < length; n++) {
        w[n] = (0.54 - 0.46 * cos(c * n)) * fft_window_gain;
    }
}
