/*
 * Copyright (c) 2014 - 2020 Honda Research Institute Japan Co., Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <stdbool.h>
#include <popt.h>
#include <math.h>
#include <sndfile.h>
#include <libharkio3.h>
#include "harktool.h"

bool hasParsedOpt = false;
void help(poptContext con, int exitCode) {
    if (hasParsedOpt)
        return;

    printf("\n");
    printf("\n");
    printf("NAME:\n");
    printf("  harktoolcli-flt2multiwav\n");
    printf("\n");
    printf("Description:\n");
    printf("  Convert the multiple raw data (32bit float wave) to a multi channel wave file.\n");
    printf("\n");
    poptPrintHelp(con, stdout, 0);
    printf("      input_raw_filenames and -l are exclusive.\n");
    printf("\n");
    printf("\n");
    exit(exitCode);
}

/*------------------------------------------------------*/
// read raw (32bit float) data
/*------------------------------------------------------*/
float* readRaw(const char* path, long *rawframelen)
{
    FILE* fp = fopen(path, "rb");
    long size, retsize;
    float *rawdata;

    if (fp == NULL) {
        harkio_Log_FileOpenFailed(path);
        return NULL;
    }

    // check file size
    fseek(fp, 0, SEEK_END);
    size = ftell(fp) / sizeof(float);
    rewind(fp);

    harkio_Log_printf(harkio_Log_Debug, "RAW file includes %ld samples.", size);

    // read files
    rawdata = (float *) malloc(size * sizeof(float));
    retsize = fread(rawdata, sizeof(float), size, fp);

    fclose(fp);
    fp = NULL;

    if (retsize != size) {
        harkio_Log_FileOpenFailed(path);
        free(rawdata);
        return NULL;
    }

    *rawframelen = size;
    return rawdata;
}

/*------------------------------------------------------*/
// save 32bit float multi-channel wav
/*------------------------------------------------------*/
int savemultiwave(int channel, const float** rawdata, int rawframelen, int samplingrate, const char *fname)
{
    SF_INFO sinfo;
    SNDFILE *sf;

    // setup header
    sinfo.frames = rawframelen;
    sinfo.channels = channel;
    sinfo.samplerate = samplingrate;
    sinfo.format = SF_FORMAT_WAV|SF_FORMAT_FLOAT;

    // format check
    if (sf_format_check(&sinfo) == 0) {
        harkio_Log_printf(harkio_Log_Error, "Format Error.");
    }

    // write data
    sf = sf_open(fname, SFM_WRITE, &sinfo);

    float buf[channel];

    for (int i = 0; i < rawframelen; i++) {
        for (int j = 0; j < channel; j++) {
            buf[j] = rawdata[j][i];
        }
        if (sf_writef_float(sf, buf, 1) != 1) {
            harkio_Log_printf(harkio_Log_Error, "sf_write_float Error. frame=%d", i);
            sf_close(sf);
            return EXIT_FAILURE;
        }
    }
    sf_close(sf);
    return EXIT_SUCCESS;
}

int main(int argc, char *argv[])
{
    // arguments
    const int samplingrate = 16000;
    char* inlistfile = NULL;
    char* logLevel = "E";

    struct poptOption optionsTable[] = {
    /*{ "longname", "shortname",
     *  argInfo,
     *  *arg, retval,
     *  description, argment description} */
    { "list", 'l',
      POPT_ARG_STRING,
      &inlistfile, 'l',
      "Input_raw_listfilename (ex. list.txt)", NULL },
    { "logLevel", '\0',
      POPT_ARG_STRING | POPT_ARGFLAG_SHOW_DEFAULT,
      &logLevel, '-',
      "Log information level", "E|W|I|D" },
    { "help", '\0',
      0,
      NULL, '?',
      "Show this help message", NULL },
    POPT_TABLEEND };

    poptContext optCon = poptGetContext(NULL, argc, (const char**)argv, optionsTable, 0);
    poptSetOtherOptionHelp(optCon, "[OPTIONS]* output_wav_filename [input_raw_filename1 filename2 ...]");

    // argc check
    if (argc <= 1) {
        harkio_Log_printf(harkio_Log_Error, "argument error!!");
        help(optCon, EXIT_FAILURE);
    }

    // option processing
    int c;
    while ((c = poptGetNextOpt(optCon)) >= 0) {
        switch (c) {
        case '-':
            if (strcmp(logLevel, "W") == 0) {
                harkio_Log_setLevel(harkio_Log_Warning);
            } else if (strcmp(logLevel, "I") == 0) {
                harkio_Log_setLevel(harkio_Log_Info);
            } else if (strcmp(logLevel, "D") == 0) {
                harkio_Log_setLevel(harkio_Log_Debug);
            } else {
                harkio_Log_setLevel(harkio_Log_Error);
            }
            hasParsedOpt = true;
            break;
        case '?':
            // help
            help(optCon, EXIT_SUCCESS);
            /* no break */
        default:
            hasParsedOpt = true;
        }
    }
    if (c < -1) {
        /* an error occurred during option processing */
        harkio_Log_printf(harkio_Log_Error, "%s: %s",
                poptBadOption(optCon, POPT_BADOPTION_NOALIAS), poptStrerror(c));
        exit(EXIT_FAILURE);
    }

    // get output file name
    const char* outfile = poptGetArg(optCon);

    if (!outfile) {
        harkio_Log_printf(harkio_Log_Error , "missing output_wav_filename");
        exit(EXIT_FAILURE);
    }

    // get input files
    char** infiles = poptGetArgs(optCon);

    if(inlistfile && infiles) {
        harkio_Log_printf(harkio_Log_Error, "%s: %s",
                poptBadOption(optCon, POPT_BADOPTION_NOALIAS), poptStrerror(POPT_ERROR_BADOPERATION));
        exit(EXIT_FAILURE);
    }

    // store input file name
    int channels = 0;
    if (inlistfile) {
        // open listfile
        FILE* fp = fopen(inlistfile, "r");
        if (fp == NULL) {
            harkio_Log_FileOpenFailed(inlistfile);
            return EXIT_FAILURE;
        }
        char infile[FILENAME_MAX] = "";
        infiles = (char**)calloc(MAX_CH, sizeof(char*));
        // read raw file list
        int i = 0;
        while(fgets(infile, sizeof(infile), fp) != NULL) {
            char* find = strchr(infile,'\n');
            if(find != NULL) *find = '\0';
            if( i >= MAX_CH ) {
                harkio_Log_printf(harkio_Log_Warning, "number of channes too large. channel max = %d", MAX_CH);
                break;
            }
            if(strlen(infile) <= 0)
                continue;
            infiles[i] = (char *)calloc(strlen(infile) + 1, sizeof(char));
            strncpy(infiles[i], infile, strlen(infile));
            infile[0] = '\0';
            i++;
        }
        channels = i;
        fclose(fp);
    } else {
        if (infiles != NULL) {
            while(infiles[channels] != NULL) {
                channels++;
            }
            if (channels > MAX_CH)
            harkio_Log_printf(harkio_Log_Warning, "number of channes too large. channel max = %d", MAX_CH);
        }
    }

    if (channels < 1) {
        harkio_Log_printf(harkio_Log_Error , "missing input_raw_files");
        exit(EXIT_FAILURE);
    }

    // raw buffers
    float* rawdatabuf[channels];
    long rawframelen[channels];
    // read input raw file
    for (int i = 0; i < channels; i++) {
        rawdatabuf[i] = readRaw(infiles[i], &rawframelen[i]);
        if(rawdatabuf[i] == NULL) {
            harkio_Log_printf(harkio_Log_Error, "Reading raw file failed.");
            for (int j = 0; j < i; j++) {
                free(rawdatabuf[j]);
            }
            if(inlistfile) {
                for (int j = 0; j < channels; j++) {
                    free(infiles[j]);
                }
                free(infiles);
            }
            exit(EXIT_FAILURE);
        }
    }

    int minframe = rawframelen[0];
    for (int i = 1; i < channels; i++) {
        if(minframe > rawframelen[i])
            minframe = rawframelen[i];
    }

    // save file
    int ret = EXIT_SUCCESS;
    if ((ret = savemultiwave(channels, rawdatabuf, minframe,
               samplingrate, outfile)) == EXIT_FAILURE) {
        harkio_Log_printf(harkio_Log_Error, "Save file failed");
    }

    // free them
    for (int j = 0; j < channels; j++) {
        free(rawdatabuf[j]);
        if(inlistfile)
            free(infiles[j]);
    }
    if(inlistfile)
        free(infiles);

    if (ret == EXIT_SUCCESS) {
        fprintf(stderr, "SUCCESS\n");
        fprintf(stderr, " output file : %s\n", outfile);
    }

    // free poptContext
    poptFreeContext(optCon);

    exit(ret);
}
