/*
 * Copyright (c) 2014 - 2020 Honda Research Institute Japan Co., Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <stdbool.h>
#include <popt.h>
#include <libharkio3.h>

bool hasParsedOpt = false;
void help(poptContext con, int exitCode) {
    if (hasParsedOpt)
        return;

    printf("\n");
    printf("\n");
    printf("NAME:\n");
    printf("  harktoolcli-mat2csv\n");
    printf("\n");
    printf("Description:\n");
    printf("  Convert csv file from mat file\n");
    printf("\n");
    poptPrintHelp(con, stdout, 0);
    printf("\n");
    printf("\n");
    exit(exitCode);
}

int main(int argc, char **argv) {
    int matrix_type;
    FILE *fp = NULL;
    COMPLEX_TYPE c_data;
    harkio_Matrix* mat;
    int isShowHeader = false;
    int isShowExp = false;
    char* logLevel = "E";

    struct poptOption optionsTable[] = {
    /*{ "longname", "shortname",
     *  argInfo,
     *  *arg, retval,
     *  description, argment description} */
    { "header", 'h',
      POPT_ARG_NONE | POPT_ARGFLAG_DOC_HIDDEN,
      &isShowHeader, 'h',
      "Show header", NULL },
    { "exp", 'e',
      POPT_ARG_NONE | POPT_ARGFLAG_DOC_HIDDEN,
      &isShowExp, 'e',
      "Show float as exponent notation", NULL },
    { "logLevel", '\0',
      POPT_ARG_STRING | POPT_ARGFLAG_SHOW_DEFAULT,
      &logLevel, '-',
      "Log information level", "E|W|I|D" },
    { "help", '\0',
      0,
      NULL, '?',
      "Show this help message", NULL },
    POPT_TABLEEND };

    poptContext optCon = poptGetContext(NULL, argc, (const char**) argv,
            optionsTable, 0);
    poptSetOtherOptionHelp(optCon,
            "[OPTIONS]* input_filename(mat) [output_filename(csv)]");

    // argc check
    if (argc <= 1) {
        harkio_Log_printf(harkio_Log_Error, "argument error!!");
        help(optCon, EXIT_FAILURE);
    }

    // option processing
    int c;
    while ((c = poptGetNextOpt(optCon)) >= 0) {
        switch (c) {
        case '-':
            if (strcmp(logLevel, "W") == 0) {
                harkio_Log_setLevel(harkio_Log_Warning);
            } else if (strcmp(logLevel, "I") == 0) {
                harkio_Log_setLevel(harkio_Log_Info);
            } else if (strcmp(logLevel, "D") == 0) {
                harkio_Log_setLevel(harkio_Log_Debug);
            } else {
                harkio_Log_setLevel(harkio_Log_Error);
            }
            hasParsedOpt = true;
            break;
        case '?':
            // help
            help(optCon, EXIT_SUCCESS);
            /* no break */
        default:
            hasParsedOpt = true;
        }
    }
    if (c < -1) {
        /* an error occurred during option processing */
        harkio_Log_printf(harkio_Log_Error, "%s: %s",
                poptBadOption(optCon, POPT_BADOPTION_NOALIAS), poptStrerror(c));
        exit(EXIT_FAILURE);
    }

    // get iput filename
    const char* infile = poptGetArg(optCon);

    if (!infile) {
        harkio_Log_printf(harkio_Log_Error, "missing input_filename");
        exit(EXIT_FAILURE);
    }

    // read mat file
    mat = harkio_Matrix_fromFile(infile);
    if (mat == NULL) {
        harkio_Log_printf(harkio_Log_Error, "input file open error!!");
        exit(EXIT_FAILURE);
    }

    matrix_type = harkio_CheckType(mat->type);
    if (matrix_type < 0) {
        harkio_Log_printf(harkio_Log_Error, "invalid matrix_type");
        exit(EXIT_FAILURE);
    }

    // get output filename
    const char* outfile = poptGetArg(optCon);

    if (outfile) {
        if ((fp = fopen(outfile, "w")) == NULL) {
            harkio_Log_printf(harkio_Log_Error, "output file open error!!");
            exit(EXIT_FAILURE);
        }
        harkio_Log_printf(harkio_Log_Debug, "output file : %s", outfile);
    } else {
        fp = stdout;
    }

    if (isShowHeader) {
        fprintf(fp, "%s,%s,%d", mat->tag, mat->type, mat->dim);
        if (mat->dim >= 1) {
            fprintf(fp, ",%d", mat->rows);
            if(mat->dim >= 2) {
                fprintf(fp, ",%d", mat->cols);
            }
        }
        fprintf(fp, "\n");
    }

    for (int row = 0; row < mat->rows; row++) {
        for (int col = 0; col < mat->cols; col++) {
            int idx = col + row * mat->cols;

            switch (matrix_type) {
            case harkio_DataType_INT32:
                fprintf(fp, "%d", mat->data.intdata[idx]);
                break;

            case harkio_DataType_FLOAT32:
                fprintf(fp, (isShowExp)?"%e":"%f", mat->data.floatdata[idx]);
                break;

            case harkio_DataType_COMPLEX:
                c_data = mat->data.complexdata[idx];
                fprintf(fp, (isShowExp)?"%e,%e":"%f,%f", crealf(c_data), cimagf(c_data));
                break;
            }

            if (col == mat->cols - 1) {
                fprintf(fp, "\n");
            } else {
                fprintf(fp, ",");
            }
        }
    }

    if (outfile) {
        fclose(fp);
    }
    fprintf(stderr, "SUCCESS\n");
    if (outfile) {
        fprintf(stderr, " output file : %s\n", outfile);
    }

    // free poptContext
    poptFreeContext(optCon);

    return EXIT_SUCCESS;
}
