/*
 * Copyright (c) 2014 - 2020 Honda Research Institute Japan Co., Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <popt.h>
#include <libharkio3.h>

bool hasParsedOpt = false;
void help(poptContext con, int exitCode) {
    if (hasParsedOpt)
        return;

    printf("\n");
    printf("\n");
    printf("NAME:\n");
    printf("  harktoolcli-mergexmls\n");
    printf("\n");
    printf("Description:\n");
    printf("   merge multiple xml files\n");
    printf("\n");
    poptPrintHelp(con, stdout, 0);
    printf("\n");
    printf("\n");
    exit(exitCode);
}

int main(int argc, char *argv[])
{

    harkio_XML* xmlsrc;
    harkio_XML* xmldst;

    char* outfile = NULL;
    FILE* fo = NULL;
    int useLast = false;
    char* logLevel = "E";

    struct poptOption optionsTable[] = {
    /*{ "longname", "shortname",
     *  argInfo,
     *  *arg, retval,
     *  description, argment description} */
    { "output", 'o',
      POPT_ARG_STRING,
      &outfile, 'o',
      "Output file name (with no output file, write result into stdout)", NULL },// use no-break-space
    { "last", 'l',
      POPT_ARG_NONE,
      &useLast, 'l',
      "Use the last argument specified when the nodes are duplicated", NULL },
    { "logLevel", '\0',
      POPT_ARG_STRING | POPT_ARGFLAG_SHOW_DEFAULT,
      &logLevel, '-',
      "Log information level", "E|W|I|D" },
    { "help", '\0',
      0,
      NULL, '?',
      "Show this help message", NULL },
    POPT_TABLEEND };

    poptContext optCon = poptGetContext(NULL, argc, (const char**)argv, optionsTable, 0);
    poptSetOtherOptionHelp(optCon, "[OPTIONS]* infile1 infile2 ...");

    // argc check
    if (argc <= 1) {
        harkio_Log_printf(harkio_Log_Error, "argument error!!");
        help(optCon, EXIT_FAILURE);
    }

    // option processing
    int c;
    while ((c = poptGetNextOpt(optCon)) >= 0) {
        switch (c) {
        case '-':
            if (strcmp(logLevel, "W") == 0) {
                harkio_Log_setLevel(harkio_Log_Warning);
            } else if (strcmp(logLevel, "I") == 0) {
                harkio_Log_setLevel(harkio_Log_Info);
            } else if (strcmp(logLevel, "D") == 0) {
                harkio_Log_setLevel(harkio_Log_Debug);
            } else {
                harkio_Log_setLevel(harkio_Log_Error);
            }
            hasParsedOpt = true;
            break;
        case '?':
            // help
            help(optCon, EXIT_SUCCESS);
            /* no break */
        default:
            hasParsedOpt = true;
        }
    }
    if (c < -1) {
        /* an error occurred during option processing */
        harkio_Log_printf(harkio_Log_Error, "%s: %s",
                poptBadOption(optCon, POPT_BADOPTION_NOALIAS), poptStrerror(c));
        exit(EXIT_FAILURE);
    }

    // get filenames
    const char** infiles = poptGetArgs(optCon);

    if (!infiles) {
        harkio_Log_printf(harkio_Log_Error , "missing infiles");
        exit(EXIT_FAILURE);
    }

    // read and assign xml data
    xmldst = harkio_XML_new();
    for (int cnt = 0; infiles[cnt] != NULL; cnt++) {
        xmlsrc = harkio_XML_newFromXML(infiles[cnt], "");

        // Positions
        if (xmlsrc->poses != NULL) {
            if (xmldst->poses != NULL) {
                if (useLast) {
                    xmldst->poses = harkio_Positions_copy(xmlsrc->poses);
                } else {
                    harkio_Log_printf(harkio_Log_Error, "<positions> Node duplicated!!");
                    exit(EXIT_FAILURE);
                }
            } else {
                xmldst->poses = harkio_Positions_copy(xmlsrc->poses);
            }
        }

        // Config
        if (xmlsrc->cfg != NULL) {
            if (xmldst->cfg != NULL) {
                if (useLast) {
                    xmldst->cfg = harkio_Config_copy(xmlsrc->cfg);
                } else {
                    harkio_Log_printf(harkio_Log_Error, "<config> Node duplicated!!");
                    exit(EXIT_FAILURE);
                }
            } else {
                xmldst->cfg = harkio_Config_copy(xmlsrc->cfg);
            }
        }

        // Neighbors
        if (xmlsrc->nbrs != NULL) {
            if (xmldst->nbrs != NULL) {
                if (useLast) {
                    xmldst->nbrs = harkio_Neighbors_copy(xmlsrc->nbrs);
                } else {
                    harkio_Log_printf(harkio_Log_Error, "<neighbors> Node duplicated!!");
                    exit(EXIT_FAILURE);
                }
            } else {
                xmldst->nbrs = harkio_Neighbors_copy(xmlsrc->nbrs);
            }
        }

        harkio_XML_delete(&xmlsrc);
    }

    // set FILE* for output
    if (outfile != NULL) {
        if ((fo = fopen(outfile , "w")) == NULL) {
            harkio_Log_printf(harkio_Log_Error , "file open error!!");
            exit(EXIT_FAILURE);
        }
    } else {
        fo = stdout;
    }

    // write xml data
    harkio_XML_write(xmldst, fo);
    harkio_XML_delete(&xmlsrc);

    if (outfile != NULL) {
        fclose(fo);
    }

    fprintf(stderr, "SUCCESS\n");
    if (outfile != NULL) {
        fprintf(stderr, " output file : %s\n",outfile);
    }

    // free poptContext
    poptFreeContext(optCon);

    return 0;
}
