/*
 * Copyright (c) 2014 - 2020 Honda Research Institute Japan Co., Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#ifndef WIN32
#define _GNU_SOURCE
#endif
#include <stdio.h>
#include <libxml/xmlreader.h>
#include <string.h>
#include <stdbool.h>
#include <popt.h>
#include <libharkio3.h>
#include "harktool.h"

bool hasParsedOpt = false;
void help(poptContext con, int exitCode) {
    if (hasParsedOpt)
        return;

    printf("\n");
    printf("\n");
    printf("NAME:\n");
    printf("  harktoolcli-validxml\n");
    printf("\n");
    printf("Description:\n");
    printf("  Consistency check of xml file\n");
    printf("\n");
    poptPrintHelp(con, stdout, 0);
    printf("\n");
    printf("    positions_type:\n");
    printf("      soundsource,tsp,impulse,microphone,noise\n");
    printf("\n");
    printf("Output :\n");
    printf("  If a xml_file is valid, the command retruns 0.\n");
    printf("\n");
    printf("\n");
    exit(exitCode);
}

int validateAsPositionXml(enum PositionType type /*MIC or SRC*/, harkio_XML* xml)
{
    char* msg = NULL;
    if (xml == NULL) {
        cat_asprintf(&msg, "[hark_xml]");
    } else {
        if (!xml->poses) {
            cat_asprintf(&msg, "[positions]");
        } else {
            char* poses_type_str = xml->poses->type ? xml->poses->type : "";
            switch(type) {
            case MIC:
                if (strcmp(poses_type_str, "microphone") != 0) {
                    cat_asprintf(&msg, "[positions (type is not microphone): %s)]", poses_type_str);
                }
                if (xml->poses && xml->poses->size <= 0) {
                    cat_asprintf(&msg , "[position is empty]");
                }
                // Mic Position id(channel) must be less than MAX_CH(64)
                for (int i = 0; i < xml->poses->size; i++) {
                    if (xml->poses->pos[i]->id >= MAX_CH) {
                        cat_asprintf(&msg, "[position (id is not less than %d: %d)]", MAX_CH, xml->poses->pos[i]->id);
                    }
                    // id must be unique
                    for (int j = 0; j < xml->poses->size; j++) {
                        if (i == j)
                            continue;
                        if (xml->poses->pos[i]->id == xml->poses->pos[j]->id) {
                            cat_asprintf(&msg, "[position (id is not unique: %d)]", xml->poses->pos[i]->id);
                        }
                    }
                }
                break;
            case SRC:
                if (strcmp(poses_type_str, "soundsource") != 0 && strcmp(poses_type_str, "tsp") != 0 && strcmp(poses_type_str, "impulse") != 0) {
                    cat_asprintf(&msg, "[positions (type is not soundsource|tsp|impulse: %s)]", poses_type_str);
                }
                break;
            case IMPULSE:
            case NOISE:
            case TSP:
                harkio_Log_printf(harkio_Log_Error, "Illegal type %d", type);
                return (EXIT_FAILURE);
            }

            // id must be unique
            for (int i = 0; i < xml->poses->size; i++) {
                for (int j = 0; j < xml->poses->size; j++) {
                    if (i == j)
                        continue;
                    if (xml->poses->pos[i]->id == xml->poses->pos[j]->id) {
                        cat_asprintf(&msg, "[position (id is not unique: %d)]", xml->poses->pos[i]->id);
                    }
                }
            }
        }
    }

    if (msg == NULL) {
        fprintf(stderr, "SUCCESS\n");
        return (EXIT_SUCCESS);
    } else {
        fprintf(stderr, "Consistency error\n");
        fprintf(stderr, "%s\n" , msg);
        free(msg);
        return (EXIT_FAILURE);
    }
}

int validate_harkxml(harkio_XML* xml)
{
    char msg[1024] = "";

    if (xml == NULL) {
        strcat(msg, "[hark_xml]");
    } else {
        if (!xml->poses) {
            strcat(msg, "[positions]");
        } else {
            enum PositionType poses_case;
            char* poses_type_str = xml->poses->type ? xml->poses->type : "";
            if (strcmp(poses_type_str ,  "soundsource") == 0) {
                poses_case = SRC;
            } else if (strcmp(poses_type_str ,  "tsp") == 0) {
                poses_case = TSP;
            } else if (strcmp(poses_type_str ,  "impulse") == 0) {
                poses_case = IMPULSE;
            } else if (strcmp(poses_type_str ,  "noise") == 0) {
                poses_case = NOISE;
            } else if (strcmp(poses_type_str ,  "microphone") == 0) {
                poses_case = MIC;
            } else {
                poses_case = -1;
                // printf("Consistency error: Unknown positions!! [%s]\n" , xml->poses->type );
                strcat(msg, "[positions (unknown type: ");
                strcat(msg, poses_type_str);
                strcat(msg, ")]");
            }

            switch (poses_case) {
            case SRC:
            case TSP:
            case IMPULSE:

                if (xml->cfg == NULL) {
                    strcat(msg ,  "[config]");
                } else {
                    if (xml->cfg->comment == NULL) {
                        strcat(msg , "[comment]");
                    }
                    if (xml->cfg->synchronousAverage < 0) {
                        strcat(msg , "[SynchronousAverage]");
                    }
                    if (xml->cfg->TSPpath      == NULL) {
                        strcat(msg , "[TSPpath]");
                    }
                    if (xml->cfg->TSPoffset    < 0) {
                        strcat(msg , "[TSPOffset]");
                    }
                    if (xml->cfg->peakSearchFrom < 0 || xml->cfg->peakSearchTo < 0) {
                        strcat(msg , "[PeakSearch]");
                    }
                    if (xml->cfg->nfft         <= 0) {
                        strcat(msg , "[nfft]");
                    }
                    if (xml->cfg->samplingRate <= 0) {
                        strcat(msg , "[samplingRate]");
                    }
                    if (xml->cfg->signalMax    <= 0) {
                        strcat(msg , "[signalMax]");
                    }
                    if (xml->cfg->TSPLength    <= 0) {
                        strcat(msg , "[TSPLength]");
                    }
                }
                /* no break */

            case NOISE:
            case MIC:
            default:
                //positions tag
                if (xml->poses && xml->poses->size <= 0) {
                    strcat(msg , "[position]");
                }
                //neighbors tag
                if (xml->nbrs    == NULL) {
                    strcat(msg , "[neighbors]");
                } else {
                    if (xml->nbrs->size <= 0) {
                        strcat(msg , "[neighbor]");
                    }
                }
            }
        }
    }

    if (strlen(msg) == 0) {
        fprintf(stderr, "SUCCESS\n");
        return (EXIT_SUCCESS);
    } else {
        fprintf(stderr, "Consistency error\n");
        fprintf(stderr, "%s\n" , msg);
        return (EXIT_FAILURE);
    }

}


int main(int argc, char **argv)
{
    int ret;
    char* logLevel = "E";
    int isMic = false;
    int isSrc = false;
    int isJson = false;

    struct poptOption optionsTable[] = {
    /*{ "longname", "shortname",
     *  argInfo,
     *  *arg, retval,
     *  description, argment description} */
    { "mic", 'm',
      POPT_ARG_NONE,
      &isMic, 'm',
      "Check for microphone positions", NULL },
    { "src", 's',
      POPT_ARG_NONE,
      &isSrc, 's',
      "Check for source (soundsource,tsp,impulse) positions", NULL },
    { "json", 'j',
      POPT_ARG_NONE | POPT_ARGFLAG_DOC_HIDDEN,
      &isJson, 'j',
      "Show info (JSON)", NULL },
    { "logLevel", '\0',
      POPT_ARG_STRING | POPT_ARGFLAG_SHOW_DEFAULT,
      &logLevel, '-',
      "Log information level", "E|W|I|D" },
    { "help", '\0',
      0,
      NULL, '?',
      "Show this help message", NULL },
    POPT_TABLEEND };

    poptContext optCon = poptGetContext(NULL, argc, (const char**)argv, optionsTable, 0);
    poptSetOtherOptionHelp(optCon, "[OPTIONS]* filename(xml)");

    // argc check
    if (argc <= 1) {
        harkio_Log_printf(harkio_Log_Error , "argument error!!");
        help(optCon, EXIT_FAILURE);
    }

    // option processing
    int c;
    while ((c = poptGetNextOpt(optCon)) >= 0) {
        switch (c) {
        case '-':
            if (strcmp(logLevel, "W") == 0) {
                harkio_Log_setLevel(harkio_Log_Warning);
            } else if (strcmp(logLevel, "I") == 0) {
                harkio_Log_setLevel(harkio_Log_Info);
            } else if (strcmp(logLevel, "D") == 0) {
                harkio_Log_setLevel(harkio_Log_Debug);
            } else {
                harkio_Log_setLevel(harkio_Log_Error);
            }
            hasParsedOpt = true;
            break;
        case '?':
            // help
            help(optCon, EXIT_SUCCESS);
            /* no break */
        default:
            hasParsedOpt = true;
        }
    }
    if (c < -1) {
        /* an error occurred during option processing */
        harkio_Log_printf(harkio_Log_Error, "%s: %s",
                poptBadOption(optCon, POPT_BADOPTION_NOALIAS), poptStrerror(c));
        exit(EXIT_FAILURE);
    }

    // get filename
    const char* filename = poptGetArg(optCon);

    if (!filename) {
        harkio_Log_printf(harkio_Log_Error , "missing filename");
        exit(EXIT_FAILURE);
    }

    harkio_XML* io3xml = harkio_XML_newFromXML(filename, "");
    if (isMic) {
        ret = validateAsPositionXml(MIC, io3xml);
    } else if (isSrc) {
        ret = validateAsPositionXml(SRC, io3xml);
    } else {
        ret = validate_harkxml(io3xml);
    }

    if (ret == EXIT_SUCCESS && isJson) {
        fprintf(stdout, "{ \"type\": \"%s\"", io3xml->poses->type);
        if (strcmp(io3xml->poses->type, "microphone") == 0) {
            fprintf(stdout, ", \"channels\": [ %d", io3xml->poses->pos[0]->id);
            for (int i = 1; i < io3xml->poses->size; i++) {
                fprintf(stdout, ",%d", io3xml->poses->pos[i]->id);
            }
            fprintf(stdout, " ]");
        } else if (io3xml->poses->pos[0]->channelsUse == 1) {
            fprintf(stdout, ", \"channels\": [ %d", io3xml->poses->pos[0]->channels[0]);
            for (int i = 1; i < io3xml->poses->pos[0]->numChannels; i++) {
                fprintf(stdout, ",%d", io3xml->poses->pos[0]->channels[i]);
            }
            fprintf(stdout, " ]");
        }
        if (io3xml->cfg) {
            fprintf(stdout, ", \"config\": { \"comment\": \"%s\"", (io3xml->cfg->comment) ? io3xml->cfg->comment : "");
            fprintf(stdout, ", \"syncadd\": %d", io3xml->cfg->synchronousAverage);
            fprintf(stdout, ", \"orgtsppath\": \"%s\"", (io3xml->cfg->TSPpath) ? io3xml->cfg->TSPpath : "");
            fprintf(stdout, ", \"tspoffset\": %d", io3xml->cfg->TSPoffset);
            fprintf(stdout, ", \"peakfrom\": %d", io3xml->cfg->peakSearchFrom);
            fprintf(stdout, ", \"peakto\": %d", io3xml->cfg->peakSearchTo);
            fprintf(stdout, ", \"fftlen\": %d", io3xml->cfg->nfft);
            fprintf(stdout, ", \"samplingrate\": %d", io3xml->cfg->samplingRate);
            fprintf(stdout, ", \"signalmax\": %d", io3xml->cfg->signalMax);
            fprintf(stdout, ", \"tsplength\": %d }", io3xml->cfg->TSPLength);
        }
        fprintf(stdout, " }\n");
    }

    // free poptContext
    poptFreeContext(optCon);
    return (ret);
}
